# Generated by Django 4.2.16 on 2024-11-04 14:50

import multiselectfield.db.fields

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structuretimers", "0005_alter_notificationrule_exclude_space_types_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notificationrule",
            name="exclude_timer_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("TF", "Theft"),
                ],
                help_text="Timer must NOT have one of the given timer types",
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="notificationrule",
            name="require_timer_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("TF", "Theft"),
                ],
                help_text="Timer must have one of the given timer types or leave blank to match any.",
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="timer",
            name="timer_type",
            field=models.CharField(
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("TF", "Theft"),
                    ("PL", "Preliminary"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
    ]
