"""Global constants."""

# pylint: disable=missing-class-docstring

from enum import IntEnum


class EveCategoryId(IntEnum):
    ORBITAL = 46
    STRUCTURE = 65


class EveGroupId(IntEnum):
    CITADEL = 1657
    CONTROL_TOWER = 365
    MERCENARY_DEN = 4810
    MOBILE_DEPOT = 1246
    PIRATE_FORWARD_OPERATING_BASE = 4644
    REFINERY = 1406
    SKYHOOK = 4736


class EveTypeId(IntEnum):
    ANSIBLEX = 35841
    ASTRAHUS = 35832
    ATHANOR = 35835
    AZBEL = 35826
    CALDARI_CONTROL_TOWER = 16213
    CALDARI_CONTROL_TOWER_MEDIUM = 20061
    CALDARI_CONTROL_TOWER_SMALL = 20062
    CUSTOMS_OFFICE = 2233
    FORTIZAR = 35833
    IHUB = 32458
    KEEPSTAR = 35834
    MERCENARY_DEN = 85230
    METENOX_MOON_DRILL = 81826
    ORBITAL_SKYHOOK = 81080
    PHAROLUX = 35840
    RAITARU = 35825
    SOTIYO = 35827
    STATION = 1529
    TATARA = 35836
    TCU = 32226
    TENEBREX = 37534
