# Generated by Django 5.2.14 on 2026-05-20 11:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structuretimers", "0006_add_theft_timer_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="timer",
            name="notification_rules",
            field=models.ManyToManyField(
                help_text="Notification rules conforming with this timer",
                through="structuretimers.ScheduledNotification",
                through_fields=("timer", "notification_rule"),
                to="structuretimers.notificationrule",
            ),
        ),
    ]
