# Generated by Django 3.2.10 on 2022-01-12 20:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("structuretimers", "0002_distances_from_staging"),
    ]

    operations = [
        migrations.AddField(
            model_name="timer",
            name="last_updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="timer",
            name="date",
            field=models.DateTimeField(
                db_index=True, help_text="Date when this timer happens", null=True
            ),
        ),
        migrations.AlterField(
            model_name="timer",
            name="timer_type",
            field=models.CharField(
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("PL", "Preliminary"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
    ]
