SET FOREIGN_KEY_CHECKS=0;
DROP TABLE IF EXISTS structuretimers_discordwebhook;
DROP TABLE IF EXISTS structuretimers_notificationrule;
DROP TABLE IF EXISTS structuretimers_notificationrule_exclude_alliances;
DROP TABLE IF EXISTS structuretimers_notificationrule_exclude_corporations;
DROP TABLE IF EXISTS structuretimers_notificationrule_require_alliances;
DROP TABLE IF EXISTS structuretimers_notificationrule_require_corporations;
DROP TABLE IF EXISTS structuretimers_schedulednotification;
DROP TABLE IF EXISTS structuretimers_timer;
SET FOREIGN_KEY_CHECKS=1;
