{% load i18n %}

<h5 for="timerboardTimer">{% translate "Timer" %}</h5>
<p>{{ object }}</p>

<h5 for="timerboardImgScreenshot">{% translate "Image" %}</h5>
<p>
    {% if object.details_image_url %}
    <a href="{{ object.details_image_url }}" target="_blank">
        <img id="timerboardImgScreenshot" class="img-responsive" alt="details image" src="{{ object.details_image_url }}"/>
    </a>
    {% else %}
    <span class="text-muted">-</span>
    {% endif %}
</p>

<h5 for="timerboardNotes">{% translate "Notes" %}</h5>
<textarea id="timerboardNotes" class="form-control" rows="3" readonly>{{ object.details_notes }}</textarea>
