{% extends "structuretimers/base.html" %}
{% load django_bootstrap5 %}
{% load i18n %}

{% block content %}
    <div class="row">
    <div class="col col-lg-8">
        <div class="card card-default">
            <div class="card-body">
                <h5 class="card-title">{% translate "Delete Structure Timer" %}</h5>
                <form action="" method="post" id="confirm-delete-form">
                    {% csrf_token %}
                    {% bootstrap_form form %}
                    <p class="card-text">
                        {% trans "Are you sure you want to delete this timer?" %}
                    </p>
                    <p class="card-text">
                        <strong>{{ object }}</strong>
                    </p>
                    <input class="btn btn-danger" type="submit" value="{% translate 'Delete' %}" />
                    {% include 'structuretimers/partials/cancel_button.html' %}
                </form>
            </div>
        </div>
    </div>
{% endblock content %}
