# Generated by Django 4.0.10 on 2023-04-18 20:36

import multiselectfield.db.fields

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("structuretimers", "0004_add_space_type_and_region_filters"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notificationrule",
            name="exclude_space_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("UN", "undefined"),
                    ("HS", "highsec"),
                    ("LS", "lowsec"),
                    ("NS", "nullsec"),
                    ("WS", "wh space"),
                ],
                help_text="Space Type must NOT be one of the selected",
                max_length=14,
            ),
        ),
        migrations.AlterField(
            model_name="notificationrule",
            name="require_space_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("UN", "undefined"),
                    ("HS", "highsec"),
                    ("LS", "lowsec"),
                    ("NS", "nullsec"),
                    ("WS", "wh space"),
                ],
                help_text="Space type must be one of the selected or leave blank to match any.",
                max_length=14,
            ),
        ),
    ]
